#include "stdafx.h"
#include "Object.h"
#include "threading.h"
#include "socket.h"
#include <crtdbg.h>

namespace seg {

	static const int TracePort = 42810;

	struct Diagnostics
	{
		struct {
			size_t AllocCount;			//mۉ(alloc)
			size_t ReallocCount;		//mۉ(realloc)
			size_t AllocSize;			//gp
			size_t AllocSizeMax;		//ő僁gp
			size_t BlockCount;			//ubN
			size_t BlockCountMax;		//őubN
		} data;

		void onAlloc( size_t size ) {
			++data.AllocCount;
			data.BlockCountMax = std::max<size_t>(data.BlockCountMax,++data.BlockCount);
			data.AllocSizeMax = std::max<size_t>(data.AllocSizeMax,(data.AllocSize+=size));
		}
		void onRealloc( size_t oldsize, size_t size ) {
			++data.ReallocCount;
			data.AllocSizeMax = std::max<size_t>(data.AllocSizeMax,(data.AllocSize+=size-oldsize));
		}
		void onFree( size_t size ) {
			--data.BlockCount;
			data.AllocSize -= size;
		}

		void init();
		void uninit();
	};
	static Diagnostics s_diagnostics;

	static void _TraceHandler( const char* message ) {
		::OutputDebugString(message);
		socket::sendto(socket::hostipaddr("localhost"),TracePort,message,strlen(message)+1);
	}
	static TraceHandler* s_tracehandler = &_TraceHandler;

	//trace handler
	DLLEXPORT void segTraceSetHandler( TraceHandler* tracehandler )
	{
		s_tracehandler = tracehandler? tracehandler: _TraceHandler;
	}

	//trace
	DLLEXPORT void segTrace( const char* message, ... )
	{
		char buf[ 1024 ] = "[seg]";
		va_list vargs;
		va_start(vargs,message);
		vsprintf(buf+5,message,vargs);
		va_end(vargs);
		strcat(buf,"\n");
		(*s_tracehandler)(buf);
	}

	//debug
	/*DLLEXPORT*/ void segDebugInit()
	{
	#if defined(_DEBUG)
		::_CrtSetDbgFlag(_CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF);
	#endif
	}
	/*DLLEXPORT*/ void segDebugUninit()
	{
		segTraceSetHandler(0);
	}

	static mutex s_memoryMutex;

	//malloc
	DLLEXPORT void* segMalloc( size_t size )
	{
		mutex::scopedlock lock(s_memoryMutex);
		s_diagnostics.onAlloc( size );
		return ::malloc(size);
	}
	//realloc
	DLLEXPORT void* segRealloc( void* ptr, size_t oldsize, size_t size )
	{
		if(!ptr){
			return segMalloc( size );
		}
		mutex::scopedlock lock(s_memoryMutex);
		s_diagnostics.onRealloc( oldsize, size );
		return ::realloc(ptr,size);
	}
	//free
	DLLEXPORT void segFree( void* ptr, size_t size )
	{
		mutex::scopedlock lock(s_memoryMutex);
		s_diagnostics.onFree( size );
		::free(ptr);
	}
	
	//malloc for squirrel
	void *segLngMalloc(SQUnsignedInteger size)
	{
		return segMalloc(size);
	}
	//realloc for squirrel
	void *segLngRealloc(void *p, SQUnsignedInteger oldsize, SQUnsignedInteger size)
	{
		return segRealloc(p,oldsize,size);
	}
	//free for squirrel
	void segLngFree(void *p, SQUnsignedInteger size)
	{
		segFree(p,size);
	}

	//IuWFNg̎QƃJE^CNg
	DLLEXPORT void segAddRef( Object* object )
	{
		incref( object );
	}
	//IuWFNgj
	DLLEXPORT void segRelease( Object* object )
	{
		decref( object );
	}

	void* Object::operator new ( size_t size )
	{
		return segMalloc( size );
	}
	void Object::operator delete ( void* ptr, size_t size )
	{
		return segFree( ptr, size );
	}

	void Diagnostics::init()
	{
		memset(&data,0,sizeof(data));

		//֐蓖
		sq_vm_malloc  = segLngMalloc;
		sq_vm_realloc = segLngRealloc;
		sq_vm_free    = segLngFree;
	}
	void Diagnostics::uninit()
	{
		segTrace("");
		segTrace(" ---- ---- ---- ----");
		segTrace("  MaxAllocSize = %4d KB",(data.AllocSizeMax+1024-1)/1024);
		segTrace(" MaxBlockCount = %4d",data.BlockCountMax);
		segTrace("    AllocCount = %4d (RE %d)",data.AllocCount+data.ReallocCount,data.ReallocCount);
		segTrace("");
		if( data.BlockCount > 0 ){
			segTrace("leak detected!!");
		}
	}

	//segŜinit/uninit
	DLLEXPORT void segInit()
	{
		socket::init();
		segDebugInit();
		s_diagnostics.init();
	}
	DLLEXPORT void segUninit()
	{
		s_diagnostics.uninit();
		segDebugUninit();
		socket::uninit();
	}

}

#pragma region global-new
struct MemoryHeader
{
	size_t size;
	size_t __pad[3];
};
void* ::operator new ( size_t size )
{
	MemoryHeader* const mh = (MemoryHeader*)seg::segMalloc(size + sizeof(MemoryHeader));
	memset(mh,0,sizeof(*mh));
	mh->size = size;
	return (mh + 1);
}
void ::operator delete ( void* ptr )
{
	MemoryHeader* const mh = (MemoryHeader*)ptr - 1;
	seg::segFree( mh, mh->size );
}
#pragma endregion
