#include "stdafx.h"
#include "Gadget.h"
#include "Context.h"

#define ctx Context::Instance()
#define SHADERSOURCE( CODE ) #CODE

namespace seg { namespace dgr {

	struct Vertex
	{
		static const int FVF = D3DFVF_XYZ | D3DFVF_DIFFUSE;

		float x;
		float y;
		float z;
		Color c;
	};

	static const char s_shaderSource[] = SHADERSOURCE(

		struct VS_INPUT
		{
			float3 pos : POSITION;
			float4 col : COLOR0;
		};
		struct VS_OUTPUT
		{
			float4 vpos : POSITION;
			float4 col : COLOR0;
		};

		float4x4 g_wvp : register(c8);

		VS_OUTPUT vsmain( VS_INPUT vsi )
		{
			VS_OUTPUT vso = (VS_OUTPUT)0;
			vso.vpos = mul(g_wvp,float4(vsi.pos,1));
			vso.col = vsi.col;
			return vso;
		}

		float4 psmain( VS_OUTPUT vso ) : COLOR0
		{
			return vso.col;
		}

		);

	static intrptr<Shader> s_shader;

	void Gadget::Init()
	{
		if(!s_shader){
			s_shader.reset(ctx.CreateShader(s_shaderSource));
		}
	}
	void Gadget::Uninit()
	{
		s_shader.reset();
	}

	void Gadget::DrawPoint( const Vector& pos, Color color )
	{
		HRESULT hr = 0;

		const Vertex vertices[] = {
			{ pos[0], pos[1], pos[2], color },
		};
		const int nbVertices = sizeof(vertices)/sizeof(vertices[0]);
		const int nbPrimitives = nbVertices / 1;

		ctx.SetShader(s_shader.get());
		hr = ctx->SetFVF( Vertex::FVF );
		hr = ctx->DrawPrimitiveUP( D3DPT_POINTLIST, nbPrimitives, vertices, sizeof(Vertex) );
		ctx.SetShader(0);
	}
	void Gadget::DrawLine( const Vector& src, const Vector& trg, Color color )
	{
		HRESULT hr = 0;

		const Vertex vertices[] = {
			{ src[0], src[1], src[2], color },
			{ trg[0], trg[1], trg[2], color },
		};
		const int nbVertices = sizeof(vertices)/sizeof(vertices[0]);
		const int nbPrimitives = nbVertices / 2;

		ctx.SetShader(s_shader.get());
		hr = ctx->SetFVF( Vertex::FVF );
		hr = ctx->DrawPrimitiveUP( D3DPT_LINELIST, nbPrimitives, vertices, sizeof(Vertex) );
		ctx.SetShader(0);
	}
	void Gadget::DrawLineStrip( const Vector* pos, int count, Color color )
	{
		HRESULT hr = 0;

		std::vector<Vertex> vertices( count );
		for( int Li = 0; Li < count; ++Li, ++pos ){
			vertices[ Li ].x = (*pos)[0];
			vertices[ Li ].y = (*pos)[1];
			vertices[ Li ].z = (*pos)[2];
			vertices[ Li ].c = color;
		}
		const int nbVertices = count;
		const int nbPrimitives = nbVertices - 1;

		ctx.SetShader(s_shader.get());
		hr = ctx->SetFVF( Vertex::FVF );
		hr = ctx->DrawPrimitiveUP( D3DPT_LINESTRIP, nbPrimitives, &vertices.front(), sizeof(Vertex) );
		ctx.SetShader(0);
	}
	
	void Gadget::DrawGrid( const Vector& center, int tile, float scale )
	{
		HRESULT hr = 0;

		const Color color[ 2 ] = { Color::Yellow, Color::White };
		const int count = (tile * 2) * (tile * 2) * 6;
		std::vector<Vertex> vertices( count );
		int index = 0;
		int adjust = 0;
		for( int Li = -tile; Li < tile; ++Li ){
			for( int Lj = -tile; Lj < tile; ++Lj, ++index ){
				const Color& c = color[(index+adjust)&1];
				Vertex* const pvx = &vertices[index*6];
				(*(Vector*)&pvx[0].x) = Vector(0.0f+Lj,0.0f,scale+Li,*(const float*)&c);
				(*(Vector*)&pvx[1].x) = Vector(scale+Lj,0.0f,scale+Li,*(const float*)&c);
				(*(Vector*)&pvx[2].x) = Vector(scale+Lj,0.0f,0.0f+Li,*(const float*)&c);
				(*(Vector*)&pvx[3].x) = (*(Vector*)&pvx[0].x);
				(*(Vector*)&pvx[4].x) = Vector(scale+Lj,0.0f,0.0f+Li,*(const float*)&c);
				(*(Vector*)&pvx[5].x) = Vector(0.0f+Lj,0.0f,0.0f+Li,*(const float*)&c);
			}
			adjust = !adjust;
		}
		const int nbVertices = count;
		const int nbPrimitives = nbVertices / 3;

		ctx.SetShader(s_shader.get());
		hr = ctx->SetFVF( Vertex::FVF );
		hr = ctx->DrawPrimitiveUP( D3DPT_TRIANGLELIST, nbPrimitives, &vertices[0], sizeof(Vertex) );
		ctx.SetShader(0);
	}

	#pragma region exports

	DLLEXPORT void segDgrGadgetDrawPoint( const Vector& pos, Color color )
	{
		Gadget::DrawPoint( pos, color );
	}
	DLLEXPORT void segDgrGadgetDrawLine( const Vector& src, const Vector& trg, Color color )
	{
		Gadget::DrawLine( src, trg, color );
	}
	DLLEXPORT void segDgrGadgetDrawLineStrip( const Vector* pos, int count, Color color )
	{
		Gadget::DrawLineStrip( pos, count, color );
	}

	#pragma endregion

}}
