#pragma once
#include "types.h"
#include <process.h>
#include <assert.h>

namespace seg {

	//! WIN32nhNX
	class Win32Handle : noncopyable
	{
	protected:
		Win32Handle() {
			handle_ = INVALID_HANDLE_VALUE;
		}
		virtual ~Win32Handle() {
			_reset();
		}

		void _reset( HANDLE h = INVALID_HANDLE_VALUE ) {
			if(_check()){
				::CloseHandle(handle_);
			}
			handle_ = h;
		}
		HANDLE _get() const {
			return handle_;
		}
		bool _check() const {
			return (handle_ != INVALID_HANDLE_VALUE);
		}

	public:
		void wait( uint timeout = INFINITE ) {
			if(_check()){
				::WaitForSingleObject(handle_,timeout);
			}
		}

	private:
		HANDLE handle_;
	};

	//! thread
	class thread : public Win32Handle
	{
	private:
		typedef uint __stdcall fn(void*);

	public:
		static void sleep( int msec ) {
			::Sleep( msec );
		}

	public:
		virtual ~thread() {
			wait();
			_reset();
		}
		void start( fn* pfn, void* arg = 0 ) {
			wait();
			uint id = 0;
			_reset((HANDLE)::_beginthreadex(0,0,pfn,arg,CREATE_SUSPENDED,&id));
			resume();
		}
		void resume() {
			if(_check()){
				::ResumeThread(_get());
			}
		}
		void suspend() {
			if(_check()){
				::SuspendThread(_get());
			}
		}
		void close() {
			wait();
			_reset();
		}
	};

	//! mutex
	class mutex : noncopyable
	{
	public:
		class scopedlock : noncopyable {
		public:
			scopedlock( mutex& _m ) : m(_m) {
				m.lock();
			}
			~scopedlock() {
				m.unlock();
			}
		private:
			mutex& m;
		};

	public:
		mutex() {
			memset(&cs,0,sizeof(cs));
			::InitializeCriticalSection(&cs);
		}
		virtual ~mutex() {
			::DeleteCriticalSection(&cs);
			memset(&cs,0,sizeof(cs));
		}
		void lock() {
			::EnterCriticalSection(&cs);
		}
	#if defined(_WIN32_WINNT) && (_WIN32_WINNT >= 0x0400)
		bool trylock() {
			return !!::TryEnterCriticalSection(&cs);
		}
	#endif
		void unlock() {
			::LeaveCriticalSection(&cs);
		}

	private:
		CRITICAL_SECTION cs;
	};

	//! semaphore
	class semaphore : public Win32Handle
	{
	public:
		explicit semaphore( int initialcount = 1, int maxcount = 1 ) {
			create(initialcount,maxcount);
		}
		void create( int initialcount = 1, int maxcount = 1 ) {
			_reset(::CreateSemaphore(0,initialcount,initialcount,0));
		}
		void signal( int count = 1 ) {
			if(_check()){
				::ReleaseSemaphore(_get(),count,0);
			}
		}
	};

	//! event
	class event : public Win32Handle
	{
	public:
		explicit event( bool initialstate = false, bool manualreset = false ) {
			create(initialstate,manualreset);
		}
		void create( bool initialstate = false, bool manualreset = false ) {
			_reset(::CreateEvent(0,manualreset,initialstate,0));
		}
		void signal() {
			if(_check()){
				::SetEvent(_get());
			}
		}
		void reset() {
			if(_check()){
				::ResetEvent(_get());
			}
		}
	};

	//! atomic int
	class atomic
	{
	public:
		atomic( int xx ) {
			x = xx;
		}
		atomic( const atomic& xx ) {
			x = xx.x;
		}
	private:
		int x;
	};
}
