#pragma once

#include <d3d9.h>
#include <d3dx9.h>
#include <assert.h>

#ifndef SEGEXPORT
#define SEGEXPORT //__declspec(dllimport)
#endif

namespace seg {

	typedef char sbyte;
	typedef unsigned char byte;
	typedef unsigned short ushort;
	typedef unsigned int uint;

	typedef char   int8;
	typedef byte   uint8;
	typedef short  int16;
	typedef ushort uint16;
	typedef int    int32;
	typedef uint   uint32;
	typedef __int64          int64;
	typedef unsigned __int64 uint64;

	class SEGEXPORT noncopyable {
	protected:
		noncopyable() {}
	private:
		noncopyable( const noncopyable& );
		void operator = ( const noncopyable& );
	};

	#pragma region bitarray
	template<int N,typename T=uint>
	class bitarray {
	public:
		typedef T element;
		static const int length = N;
		static const int component_size = sizeof(element)*8;
		static const element mask = 1UL << (component_size - 1);
		static const int element_size = (length+component_size-1)/component_size;
	public:
		bitarray() {
			clear();
		}
		void clear() {
			memset(b,0,sizeof(b));
		}
		void set( int index ) {
			assert(0 <= index && index < length);
			b[index/component_size] |= mask >> (index % component_size);
		}
		void reset( int index ) {
			assert(0 <= index && index < length);
			b[index/component_size] &= ~((T)(mask >> (index % component_size)));
		}
		void set( int index, bool flag ) {
			flag? set(index): reset(index);
		}
		bool get( int index ) const {
			assert(0 <= index && index < length);
			return !!(b[index/component_size] & (mask >> (index % component_size)));
		}
		bool operator [] ( int index ) const {
			return get( index );
		}
	private:
		T b[element_size];
	};
	typedef bitarray<sizeof(uint8)*8,byte>    bit8;
	typedef bitarray<sizeof(uint16)*8,ushort> bit16;
	typedef bitarray<sizeof(uint32)*8,uint32> bit32;
	typedef bitarray<sizeof(uint64)*8,uint64> bit64;

	template<int N,typename T>
	class bitstack
	{
	public:
		bitstack() {
			n = 0;
		}
		bool empty() const {
			return (n == 0);
		}
		int length() const {
			return n;
		}
		void clear() {
			b.clear();
			n = 0;
		}
		int push( uint bits, int len = 1 ) {
			if(n + len > N){
				return -1;
			}
			const uint mask = 1<<(len-1);
			for(;len>0;--len,bits<<=1){
				b.set(n++,!!(bits & mask));
			}
			return 0;
		}
		uint pop( int len = 1 ) {
			assert(len <= n);
			uint bits = 0;
			for(int Li=0;Li<len;++Li){
				bits |= b[--n] << Li;
			}
			return bits;
		}
	private:
		bitarray<N,T> b;
		int n;
	};

	template<int N,typename T>
	class bitqueue
	{
	public:
		bitqueue() {
			s = e = 0;
		}
		bool empty() const {
			return (s == e);
		}
		int length() const {
			return (e - s);
		}
		void clear() {
			b.clear();
			s = e = 0;
		}
		int push( uint bits, int len = 1 ) {
			if( e + len - s > N ){
				return -1;
			}
			const uint mask = 1<<(len-1);
			for(;len>0;--len,bits<<=1){
				b.set(e++%N,!!(bits & mask));
			}
			return 0;
		}
		uint pop( int len = 1 ) {
			if(e - s < len){
				return 0;
			}
			uint bits = 0;
			for(;len>0;--len){
				bits = (b[s++%N]?1:0) | (bits << 1);
			}
			if(s >= N){
				s -= N;
				e -= N;
			}
			return bits;
		}
	private:
		bitarray<N,T> b;
		short s;
		short e;
	};
	#pragma endregion

	class SEGEXPORT Id {
	private:
		struct raw_t {
			uint b[ 4 ];
		};
		static int encode( raw_t* raw, const char* str, const char*const* cp );
		static int decode( char* buf, size_t len, const raw_t& raw, const char*const* cp );
	public:
		explicit Id( const char* str = "", const char*const* cp = 0 ) {
			reset(str,cp);
		}
		int reset( const char* str, const char*const* cp = 0 ) {
			return encode(&raw,str,cp);
		}
		char* get( char* buf, size_t len, const char*const* cp = 0 ) const {
			return decode(buf,len,raw,cp)? "": buf;
		}
	private:
		raw_t raw;
	};
	
	static const float PI = 3.1415926535f;
	static inline float DegToRad( float degrees ) {
		return degrees * PI / 180.0f;
	}
	static inline float RadToDeg( float radians ) {
		return radians * 180.0f / PI;
	}

	#pragma region Color
	class Color
	{
	public:
		static const Color Black;
		static const Color White;
		static const Color Gray;
		static const Color Red;
		static const Color Green;
		static const Color Blue;
		static const Color Yellow;
		static const Color Cyan;

	public:
		explicit Color() {
			c = D3DCOLOR_XRGB(255,255,255);
		}
		explicit Color( byte x ) {
			c = D3DCOLOR_XRGB(x,x,x);
		}
		explicit Color( byte r, byte g, byte b, byte a = 255 ) {
			c = D3DCOLOR_RGBA(r,g,b,a);
		}

	private:
		uint c;
	};
	#pragma endregion

	#pragma region Vector
	__declspec(align(16))
	class Vector
	{
	public:
		static const Vector Zero;
		static const Vector XUnit;
		static const Vector YUnit;
		static const Vector ZUnit;
		static const Vector WUnit;

	private:
		explicit Vector( const D3DXVECTOR3& vv, float w ) {
			v[0] = vv[0];
			v[1] = vv[1];
			v[2] = vv[2];
			v[3] = w;
		}
		explicit Vector( const D3DXVECTOR4& vv ) {
			v = vv;
		}
	public:
		explicit Vector() {
			v = Zero.v;
		}
		explicit Vector( float x ) {
			v = D3DXVECTOR4(x,x,x,x);
		}
		explicit Vector( float x, float y, float z = 0.0f, float w = 1.0f ) {
			v = D3DXVECTOR4(x,y,z,w);
		}

		float operator[] ( int n ) const {
			return v[n];
		}
		
		Vector operator + () const {
			return Vector(v);
		}
		Vector operator - () const {
			return Vector(-v);
		}
		Vector operator + ( const Vector& rhs ) const {
			return Vector(v + rhs.v);
		}
		Vector& operator += ( const Vector& rhs ) {
			return *(Vector*)&(v += rhs.v);
		}
		Vector operator - ( const Vector& rhs ) const {
			return Vector(v - rhs.v);
		}
		Vector& operator -= ( const Vector& rhs ) {
			return *(Vector*)&(v -= rhs.v);
		}

		float Dot3( const Vector& rhs ) const {
			return D3DXVec3Dot((const D3DXVECTOR3*)&v,(const D3DXVECTOR3*)&rhs.v);
		}
		float Dot( const Vector& rhs ) const {
			return D3DXVec4Dot(&v,&rhs.v);
		}
		float Length3() const {
			return D3DXVec3Length((const D3DXVECTOR3*)&v);
		}
		float Length() const {
			return D3DXVec4Length(&v);
		}
		Vector Normalize3() const {
			D3DXVECTOR4 tmp = WUnit.v;
			return *(Vector*)D3DXVec3Normalize((D3DXVECTOR3*)&tmp,(const D3DXVECTOR3*)&v);
		}
		Vector Normalize() const {
			D3DXVECTOR4 tmp = WUnit.v;
			return *(Vector*)D3DXVec4Normalize(&tmp,&v);
		}
		Vector Cross( const Vector& rhs ) const {
			D3DXVECTOR4 tmp = WUnit.v;
			return *(Vector*)D3DXVec3Cross((D3DXVECTOR3*)&tmp,(const D3DXVECTOR3*)&v,(const D3DXVECTOR3*)&rhs.v);
		}

	private:
		D3DXVECTOR4 v;

		friend class Matrix;
	};
	#pragma endregion

	#pragma region Matrix
	__declspec(align(16))
	class Matrix
	{
	public:
		static const Matrix Zero;
		static const Matrix Identity;

		static Matrix Perspective( float fovy_radians, float aspect, float znear, float zfar ) {
			D3DXMATRIX tmp;
			return *(Matrix*)D3DXMatrixPerspectiveFovRH(&tmp,fovy_radians,aspect,znear,zfar);
		}
		static Matrix LookAt( const Vector& eye, const Vector& aim, const Vector& upv ) {
			D3DXMATRIX tmp;
			return *(Matrix*)D3DXMatrixLookAtRH(&tmp,(const D3DXVECTOR3*)&eye,(const D3DXVECTOR3*)&aim,(const D3DXVECTOR3*)&upv);
		}
		static Matrix Scaling( const Vector& vec ) {
			D3DXMATRIX tmp;
			return *(Matrix*)D3DXMatrixScaling(&tmp,vec[0],vec[1],vec[2]);
		}
		static Matrix Translation( const Vector& vec ) {
			D3DXMATRIX tmp;
			return *(Matrix*)D3DXMatrixTranslation(&tmp,vec[0],vec[1],vec[2]);
		}
		static Matrix RotationX( float radians ) {
			D3DXMATRIX tmp;
			return *(Matrix*)D3DXMatrixRotationX(&tmp,radians);
		}
		static Matrix RotationY( float radians ) {
			D3DXMATRIX tmp;
			return *(Matrix*)D3DXMatrixRotationY(&tmp,radians);
		}
		static Matrix RotationZ( float radians ) {
			D3DXMATRIX tmp;
			return *(Matrix*)D3DXMatrixRotationZ(&tmp,radians);
		}
		static Matrix RotationAxis( const Vector& axis, float radians ) {
			D3DXMATRIX tmp;
			return *(Matrix*)D3DXMatrixRotationAxis(&tmp,(D3DXVECTOR3*)&axis,radians);
		}

	public:
		explicit Matrix() {
			m = Identity.m;
		}
		explicit Matrix( const Vector& c0, const Vector& c1, const Vector& c2, const Vector& c3 ) {
			*(Vector*)&m.m[0] = c0;
			*(Vector*)&m.m[1] = c1;
			*(Vector*)&m.m[2] = c2;
			*(Vector*)&m.m[3] = c3;
		}

		const Vector& operator[] ( int n ) const {
			return *(const Vector*)&m.m[n];
		}

		Matrix& operator *= ( const Matrix& rhs ) {
			return *(Matrix*)D3DXMatrixMultiply(&m,&rhs.m,&m);
		}
		Matrix operator * ( const Matrix& rhs ) const {
			D3DXMATRIX tmp;
			return *(Matrix*)D3DXMatrixMultiply(&tmp,&rhs.m,&m);
		}

		Matrix& Transpose() {
			return *(Matrix*)D3DXMatrixTranspose(&m,&m);
		}
		Matrix& Inverse() {
			float det;
			return *(Matrix*)D3DXMatrixInverse(&m,&det,&m);
		}
		Vector Transform( const Vector& vec ) const {
			D3DXVECTOR4 tmp;
			return *(Vector*)D3DXVec4Transform(&tmp,&vec.v,&m);
		}
		Vector TransformCoord( const Vector& vec ) const {
			D3DXVECTOR4 tmp(0.0f,0.0f,0.0f,1.0f);
			return *(Vector*)D3DXVec3TransformCoord((D3DXVECTOR3*)&tmp,(const D3DXVECTOR3*)&vec.v,&m);
		}
		Vector TransformNormal( const Vector& vec ) const {
			D3DXVECTOR4 tmp(0.0f,0.0f,0.0f,1.0f);
			return *(Vector*)D3DXVec3TransformNormal((D3DXVECTOR3*)&tmp,(const D3DXVECTOR3*)&vec.v,&m);
		}

	private:
		D3DXMATRIX m;
	};
	#pragma endregion

	#pragma region AABB
	__declspec(align(16))
	class AABB
	{
	public:
		static const AABB Invalid;
		static const AABB Zero;
		static const AABB Identity;

	public:
		explicit AABB() {
			v[0] = Invalid.v[0];
			v[1] = Invalid.v[1];
		}
		explicit AABB( const Vector& vec ) {
			v[0] = v[1] = vec;
		}
		explicit AABB( const Vector& vmin, const Vector& vmax ) {
			v[0] = vmin;
			v[1] = vmax;
		}

		const Vector& operator[] ( int n ) const {
			return *(const Vector*)&v[n];
		}

	private:
		Vector v[2];//{min,max}
	};
	#pragma endregion

}

#include "seg/intrptr.h"
