//	shive's code 2008.08.22
//=============================================================================

#include "stdafx.h"
#include "Allocator.h"
#include "../com/debug.h"

#undef new
#undef malloc

#ifdef MY_DEBUG
//#	define MEMORY_TRACE
#	define MEMORY_LEAKCHECK
#	define MEMORY_FILL
#endif
#ifdef MEMORY_TRACE
#	define WRITELINE( ... )	debug::writeline( __VA_ARGS__ );
#else
#	define WRITELINE( ... )
#endif

namespace shive { namespace mem {

	using com::debug;

	static ushort s_allocateCount = 0;
	static ushort s_breakIndex = -1;

	static inline uint calc1k( uint value )
	{
		return (value + 1024 - 1) / 1024;
	}

	class Heap::Impl : public systemheap
	{
	public:
	#ifdef MEMORY_LEAKCHECK
		struct DebugChain {
			DebugChain* prev;
			DebugChain* next;
		};
		struct DebugInfo {
			DebugChain chain;
			const char* file;
			ushort line;
			ushort allocateIndex;
		};
	#else
		typedef void DebugInfo;
	#endif
		struct MBH {
			Impl* heap;
			uint size;
			uint state;
			DebugInfo* debug;
		};

	public:
		explicit Impl( const char* heapname )
		{
			memset(&this->info_,0,sizeof(this->info_));
			strcpy_s(this->info_.Tag,heapname);
		#ifdef MEMORY_LEAKCHECK
			memset(&this->chain_,0,sizeof(this->chain_));
		#endif
			this->nextheap_ = 0;
		}
		~Impl()
		{
			com::mutex::scopedlock lock(this->mutex_);
		#ifdef MEMORY_LEAKCHECK
			DebugChain* iter = this->chain_.next;
			if( iter ){
				do {
					DebugInfo* const debug = (DebugInfo*)iter;
					iter = iter->next;
					MBH* const mbh = (MBH*)(debug + 1);
					debug::writeline("%s(%d): memleak[ %d ]: 0x%08X ( %d B) //%s",
						debug->file,debug->line,
						debug->allocateIndex,mbh+1,mbh->size,
						this->info_.Tag);
					SYSTEMHEAP_FREE( debug );
				} while( iter );
			}
		#endif
			if( this->info_.MaxAllocatedBlocks > 0 ){
				debug::writeline(
					"mem: maxblocks = %4d / maxsize = %4d KB    //%s",
					this->info_.MaxAllocatedBlocks,
					calc1k(this->info_.MaxAllocatedSize),
					this->info_.Tag );
			}
		}
		//fobO񂠂
		void* Allocate( size_t size, int boundary, const char* file, int line )
		{
			com::mutex::scopedlock lock(this->mutex_);

		#ifdef MEMORY_LEAKCHECK
			const int header = sizeof(MBH) + sizeof(DebugInfo);
		#else
			const int header = sizeof(MBH);
		#endif
			void* ptr = SYSTEMHEAP_MALLOC( size + header, boundary, header );
			assert(ptr);
			memset(ptr,0,header);
			ptr = (void*)((uintptr_t)ptr + header);

			MBH* const mbh = (MBH*)((uintptr_t)ptr - sizeof(MBH));
			mbh->heap = this;
			mbh->size = size;
		#ifdef MEMORY_LEAKCHECK
			DebugInfo* const debug = (DebugInfo*)((uintptr_t)mbh - sizeof(DebugInfo));
			mbh->debug = debug;
			debug->chain.prev = &this->chain_;
			debug->chain.next = this->chain_.next;
			if( this->chain_.next ){
				this->chain_.next->prev = &debug->chain;
			}
			this->chain_.next = &debug->chain;
			debug->file = file;
			debug->line = line;
			debug->allocateIndex = s_allocateCount;
		#endif
		#ifdef MEMORY_FILL
			memset(ptr,0xDD,size);
		#endif
			
			WRITELINE("mem: +0x%08X ( %d B) //%s",(uintptr_t)ptr,size,this->info_.Tag);

			this->info_.AllocatedBlocks++;
			this->info_.AllocatedSize += size;
			if( this->info_.MaxAllocatedBlocks < this->info_.AllocatedBlocks ){
				this->info_.MaxAllocatedBlocks = this->info_.AllocatedBlocks;
			}
			if( this->info_.MaxAllocatedSize < this->info_.AllocatedSize ){
				this->info_.MaxAllocatedSize = this->info_.AllocatedSize;
			}

			return ptr;
		}
		void Deallocate( void* ptr, size_t )
		{
			com::mutex::scopedlock lock(this->mutex_);
			MBH* const mbh = (MBH*)((uintptr_t)ptr - sizeof(MBH));
		#ifdef MEMORY_LEAKCHECK
			DebugInfo* const debug = mbh->debug;
			if( debug ){
				DebugChain* const prev = debug->chain.prev? debug->chain.prev: &this->chain_;
				prev->next = debug->chain.next;
				if( prev->next ){
					prev->next->prev = prev;
				}
				debug->chain.prev = debug->chain.next = 0;
			}
		#endif
			WRITELINE("mem: -0x%08X ( %d B) //%s",(uintptr_t)ptr,mbh->size,this->info_.Tag);
			this->info_.AllocatedBlocks--;
			this->info_.AllocatedSize -= mbh->size;
		#ifdef MEMORY_FILL
			memset(ptr,0xEE,mbh->size);
		#endif
		#ifdef MEMORY_LEAKCHECK
			SYSTEMHEAP_FREE( debug );
		#else
			SYSTEMHEAP_FREE( mbh );
		#endif
		}
		const HeapInfo& GetHeapInfo() const
		{
			com::mutex::scopedlock lock(this->mutex_);
			return this->info_;
		}
	private:
		HeapInfo info_;
	#ifdef MEMORY_LEAKCHECK
		DebugChain chain_;
	#endif
		Heap::Impl* nextheap_;
		mutable com::mutex mutex_;
	};
	__declspec(thread) static Heap* s_currentHeap = 0;
	static com::mutex s_changeHeapMutex;//Jgq[vύX邽߂̓IuWFNg
	static objptr<Heap> s_scriptHeap;
	static objptr<Heap> s_globalHeap;

	static inline Heap* getcurrentheap()
	{
		return s_currentHeap? s_currentHeap: s_globalHeap.get();
	}

	void Allocator::Initialize()
	{
		com::mutex::scopedlock lock(s_changeHeapMutex);
		s_globalHeap.reset( new Heap("GLOBAL") );
		s_scriptHeap.reset( new Heap("SCRIPT") );
	}
	void Allocator::Terminate()
	{
		com::mutex::scopedlock lock(s_changeHeapMutex);
		assert(!s_currentHeap);
		s_scriptHeap.reset();
		s_globalHeap.reset();
	}

	void Allocator::SetAllocBreak( int index )
	{
		s_breakIndex = index;
	}
	void* Allocator::Allocate( size_t size, int boundary, const char* file, int line )
	{
		com::mutex::scopedlock lock(s_changeHeapMutex);
		//ŒmۃTCY
		size = (size + 0x10 - 1) & ~(size_t)0xF;
		if( size < 0x10 ){
			size = 0x10;
		}
		//ŒACgTCY
		if( boundary < 0x10 ){
			boundary = 0x10;
		}
		void* const ptr = getcurrentheap()->pimpl->Allocate( size, boundary, file, line );
		if( s_allocateCount == s_breakIndex ){
			__debugbreak();
		}
		++s_allocateCount;
		return ptr;
	}
	void Allocator::Deallocate( void* ptr, size_t size )
	{
		if( ptr ){
			Heap::Impl::MBH* const mbh = (Heap::Impl::MBH*)ptr - 1;
			assert(mbh->heap);
			mbh->heap->Deallocate( ptr, size );
		}
	}
	const HeapInfo& Allocator::GetCurrentHeapInfo()
	{
		com::mutex::scopedlock lock(s_changeHeapMutex);
		return getcurrentheap()->GetHeapInfo();
	}

	Heap::Heap( const char* tag ) : pimpl( new Impl( tag ) )
	{
	}
	Heap::~Heap()
	{
		assert(s_currentHeap != this);//JĝƂ͎˂Ȃ
		delete pimpl;
	}
	const HeapInfo& Heap::GetHeapInfo() const
	{
		return pimpl->GetHeapInfo();
	}
	//q[vgpJnAȑÕq[vԂ
	Heap* Heap::activate_()
	{
		com::mutex::scopedlock lock(s_changeHeapMutex);
		Heap* old = getcurrentheap();
		s_currentHeap = (this == s_globalHeap)? 0: this;
		return old;
	}

}}

namespace shive {

	void* malloc( size_t size, const char* file, int line )
	{
		return mem::Allocator::Allocate( size, 0x10, file, line );
	}
	void* realloc( void* ptr, size_t oldsize, size_t size, const char* file, int line )
	{
		if( ptr && oldsize >= size ){
			return ptr;
		}
		void* renewal = malloc( size, file, line );
		if( ptr ){
			memcpy( renewal, ptr, size );
		}
		free( ptr, oldsize );
		return renewal;
	}
	void free( void* ptr, size_t size )
	{
		mem::Allocator::Deallocate( ptr, size );
	}

	namespace scp {

		void* malloc( size_t size, const char* file, int line )
		{
			mem::Layer layer(mem::s_scriptHeap.get());
			return ::shive::malloc( size, file, line );
		}
		void* realloc( void* ptr, size_t oldsize, size_t size, const char* file, int line )
		{
			mem::Layer layer(mem::s_scriptHeap.get());
			return ::shive::realloc( ptr, oldsize, size, file, line );
		}
		void free( void* ptr, size_t size )
		{
			::shive::free( ptr, size );
		}

	}

}
